import ast
from typing import List

class VariableFinder(ast.NodeVisitor):
    def __init__(self):
        self.variables = set()

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            self.variables.add(node.id)


import random
import time
from typing import List

from testgen.models.test_case import TestCase
from testgen.reinforcement.environment import ReinforcementEnvironment
from testgen.service.logging_service import get_logger


class ReinforcementAgent:
    def __init__(self, file_name: str, environment: ReinforcementEnvironment, test_cases: List[TestCase], q_table=None):
        self.learning_rate = 0.1
        self.file_name = file_name
        self.env = environment
        self.q_table = q_table if q_table else {}
        self.actions = ["add", "merge", "remove", "z3"]
        self.logger = get_logger()

    def collect_test_cases(self) -> List[TestCase]:
        max_time = 30
        if not self.q_table:
            print("Q_TABLE IS EMPTY, RUN TRAIN FIRST")
            return []
        else:
            current_state = self.env.get_state()

            goal_state: float = 100.0

            start_time = time.time()

            while current_state[0] != goal_state and time.time() - start_time < max_time:
                action = self.choose_action(current_state)
                next_state, reward = self.env.step(action)

                if not isinstance(next_state, tuple) or len(next_state) != 2:
                    raise ValueError(f"Expected new_state to be a tuple (covered_statements, coverage_percentage, len(test_cases)), but got: {next_state}")

                self.update_q_table(next_state, action, next_state, reward)

                current_state = next_state

            return self.env.test_cases


    def do_q_learning(self, episodes=10):
        max_time = 30
        best_coverage = 0.0
        best_test_cases = []

        for episode in range(episodes):
            print(f"\nNEW EPISODE {episode}")
            self.env.reset()

            current_state = self.env.get_state()
            self.logger.debug(f"Current state after reset: {current_state}")

            goal_state: float = 100.0
            steps_in_episode = 1
            max_steps_per_episode = 100

            start_time = time.time()

            while current_state[0] != goal_state and steps_in_episode < max_steps_per_episode and time.time() - start_time < max_time:
                print(f"\nStep {steps_in_episode} in episode {episode}")

                action = self.choose_action(current_state)
                next_state, reward = self.env.step(action)

                if not isinstance(next_state, tuple) or len(next_state) != 2:
                    raise ValueError(f"Expected new_state to be a tuple (covered_statements, coverage_percentage, len(test_cases)), but got: {next_state}")

                print(f"AFTER NEW STATE, REWARD: {reward}")

                # Update q_table
                self.update_q_table(current_state, action, next_state, reward)
                current_state = next_state

                steps_in_episode += 1
                if current_state[0] > best_coverage:
                    best_coverage = current_state[0]
                    best_test_cases = self.env.test_cases.copy()
                    self.logger.debug(f"New best coverage: {best_coverage}% with {len(best_test_cases)} test cases")
                elif current_state[0] == best_coverage and len(best_test_cases) > len(self.env.test_cases):
                    best_test_cases = self.env.test_cases.copy()
                    self.logger.debug(f"New best coverage: {best_coverage}% with {len(best_test_cases)} test cases")

        return best_test_cases


    def choose_action(self, state):
        EXPLORATION = 0
        EXPLOITATION = 1

        weights = [0.33, 0.67]
        
        choice = random.choices([EXPLORATION, EXPLOITATION], weights=weights, k=1)[0]
        action_list = self.get_action_list(state[1])

        if not isinstance(state, tuple) or len(state) != 2:
            raise ValueError(f"Expected state to be a tuple (covered_statements, coverage_percentage, len(test_cases)), but got: {state}")

        if choice == EXPLORATION:
            chosen_action = random.choice(action_list)
            print(f"CHOSEN EXPLORATION ACTION: {chosen_action}")
            return chosen_action
        else:
            chosen_action = max(action_list, key=lambda action: self.q_table.get((state, action), 0), default=random.choice(action_list))
            print(f"CHOSEN EXPLOITATION ACTION: {chosen_action}")
            return chosen_action

    

    @staticmethod
    def get_action_list(test_case_length: int) -> List[str]:
        action_list = ["add", "z3"]
        if test_case_length >= 2:
            action_list.extend(["merge", "remove"])
        return action_list

    def update_q_table(self, state: tuple, action: str, new_state:tuple, reward:float):
        current_q = self.q_table.get((state, action), 0)
        self.logger.debug(f"CURRENT Q: {current_q}")
        valid_actions = self.get_action_list(new_state[1])

        max_next_q = max(self.q_table.get((new_state, a), 0) for a in valid_actions)
        self.logger.debug(f"MAX NEXT Q: {max_next_q}")

        print(f"UPDATING Q TABLE FOR STATE: {state}, ACTION: {action} WITH REWARD: {reward}")
        new_q = (1 - self.learning_rate) * current_q + self.learning_rate * (reward + max_next_q)
        
        self.q_table[(state, action)] = new_q
        
        """def optimize_test_suit(self, current_state, executable_statements):
        # Try to optimize test cases by repeatedly performing remove actions if reached full coverage
        test_case_count = current_state[1]
        optimization_attempts = min(10, test_case_count - 1)

        for _ in range(optimization_attempts):
            if test_case_count <= 1:
                break

            action = "remove"
            next_state, reward = self.env.step(action)

            new_covered = next_state[0]
            new_uncovered = [stmt for stmt in executable_statements if stmt not in new_covered]

            if len(new_uncovered) == 0:
                current_state = next_state
                test_case_count = current_state[2]
                print(f"Optimized to {test_case_count} test cases.")
            else:
                # Add a test case back if removing broke coverage
                self.env.step("add")
                break

        return current_state"""


def bin_and(a: bool, b: bool):
    if a == True:
        if b == True:
            return True
        else:
            return False
    else:
        if b == True:
            return False
        else:
            return False

def bin_xor(a: bool, b: bool):
    if a == True:
        if b == True:
            return False
        else:
            return True
    else:
        if b == True:
            return True
        else:
            return False

def status_flags(active: bool, verified: bool, admin: bool):
    if active == True:
        if verified == True:
            if admin == True:
                return 'admin-verified'
            else:
                return 'user-verified'
        else:
            if admin == True:
                return 'admin-unverified'
            else:
                return 'user-unverified'
    else:
        if verified == True:
            if admin == True:
                return 'admin-verified'
            else:
                return 'inactive'
        else:
            if admin == True:
                return 'admin-unverified'
            else:
                return 'inactive'




import os
import sqlite3
import json
import time
import ast
from datetime import datetime
from typing import List, Tuple

from testgen.models.test_case import TestCase
from testgen.sqlite.db import create_database

class DBService:
    def __init__(self, db_name="testgen.db"):
        self.db_name = db_name
        self.conn = None
        self.cursor = None
        self._connect()

    def _connect(self):
        """Establish connection to the database."""
        if not os.path.exists(self.db_name):
            create_database(self.db_name)
        
        self.conn = sqlite3.connect(self.db_name)
        self.conn.row_factory = sqlite3.Row
        self.cursor = self.conn.cursor()
        # Enable foreign keys
        self.cursor.execute("PRAGMA foreign_keys = ON;")

    def close(self):
        """Close the database connection."""
        if self.conn:
            self.conn.close()
            self.conn = None
            self.cursor = None

    def insert_test_suite(self, name: str) -> int:
        """Insert a test suite and return its ID."""
        self.cursor.execute(
            "INSERT INTO TestSuite (name, creation_date) VALUES (?, ?)",
            (name, datetime.now())
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_source_file(self, path: str, lines_of_code: int) -> int:
        """Insert a source file and return its ID."""
        # Check if file already exists
        self.cursor.execute("SELECT id FROM SourceFile WHERE path = ?", (path,))
        existing = self.cursor.fetchone()
        if existing:
            return existing[0]
        
        self.cursor.execute(
            "INSERT INTO SourceFile (path, lines_of_code, last_modified) VALUES (?, ?, ?)",
            (path, lines_of_code, datetime.now())
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_function(self, name: str, start_line: int, end_line: int, source_file_id: int) -> int:
        """Insert a function and return its ID."""
        num_lines = end_line - start_line + 1
        
        # Check if function already exists for this source file
        self.cursor.execute(
            "SELECT id FROM Function WHERE name = ? AND source_file_id = ?", 
            (name, source_file_id)
        )
        existing = self.cursor.fetchone()
        if existing:
            return existing[0]
            
        self.cursor.execute(
            "INSERT INTO Function (name, start_line, end_line, num_lines, source_file_id) VALUES (?, ?, ?, ?, ?)",
            (name, start_line, end_line, num_lines, source_file_id)
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_test_case(self, test_case: TestCase, test_suite_id: int, function_id: int, test_method_type: int) -> int:
        """Insert a test case and return its ID."""
        # Convert inputs and expected output to JSON strings
        inputs_json = json.dumps(test_case.inputs)
        expected_json = json.dumps(test_case.expected)
        
        self.cursor.execute(
            "INSERT INTO TestCase (name, expected_output, input, test_function, last_run_time, test_method_type, test_suite_id, function_id) "
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
            (
                f"test_{test_case.func_name}",
                expected_json,
                inputs_json,
                test_case.func_name,
                datetime.now(),
                test_method_type,
                test_suite_id,
                function_id
            )
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_test_result(self, test_case_id: int, status: bool, error: str = None) -> int:
        """Insert a test result and return its ID."""
        self.cursor.execute(
            "INSERT INTO TestResult (test_case_id, status, error, execution_time) VALUES (?, ?, ?, ?)",
            (test_case_id, status, error, datetime.now())
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_coverage_data(self, file_name: str, executed_lines: str, missed_lines: str,
                           branch_coverage: float, source_file_id: int) -> int:
        """Insert coverage data and return its ID."""
        self.cursor.execute(
            "INSERT INTO CoverageData (file_name, executed_lines, missed_lines, branch_coverage, source_file_id) "
            "VALUES (?, ?, ?, ?, ?)",
            (file_name, executed_lines, missed_lines, branch_coverage, source_file_id)
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def save_test_generation_data(self, file_path: str, test_cases: List[TestCase], 
                                test_method_type: int, class_name: str = None) -> Tuple[int, List[int]]:
        """
        Save all data related to a test generation run.
        Returns the test suite ID and a list of test case IDs.
        """
        # Count lines in the source file
        with open(file_path, 'r') as f:
            lines_of_code = len(f.readlines())
        
        # Create test suite
        strategy_names = {1: "AST", 2: "Fuzz", 3: "Random", 4: "Reinforcement"}
        suite_name = f"{strategy_names.get(test_method_type, 'Unknown')}_Suite_{int(time.time())}"
        test_suite_id = self.insert_test_suite(suite_name)
        
        # Insert source file
        source_file_id = self.insert_source_file(file_path, lines_of_code)
        
        # Process functions and test cases
        test_case_ids = []
        function_ids = {}  # Cache function IDs to avoid redundant queries
        
        for test_case in test_cases:
            # Extract function name from test case
            func_name = test_case.func_name
            
            # If function not already processed
            if func_name not in function_ids:
                # Get function line numbers
                start_line, end_line = self._get_function_line_numbers(file_path, func_name)
                function_id = self.insert_function(func_name, start_line, end_line, source_file_id)
                function_ids[func_name] = function_id
            
            # Insert test case
            test_case_id = self.insert_test_case(
                test_case, 
                test_suite_id, 
                function_ids[func_name], 
                test_method_type
            )
            test_case_ids.append(test_case_id)
        
        return test_suite_id, test_case_ids
    
    def _get_function_line_numbers(self, file_path: str, function_name: str) -> Tuple[int, int]:
        """
        Extract the start and end line numbers for a function in a file.
        Returns a tuple of (start_line, end_line).
        """
        try:
            # Load the file and parse it
            with open(file_path, 'r') as f:
                file_content = f.read()
                
            tree = ast.parse(file_content)
            
            # Find the function definition
            for node in ast.walk(tree):
                if isinstance(node, ast.FunctionDef) and node.name == function_name:
                    end_line = node.end_lineno if hasattr(node, 'end_lineno') else node.lineno + 5  # Estimate if end_lineno not available
                    return node.lineno, end_line
                    
            # Also look for class methods
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    for class_node in node.body:
                        if isinstance(class_node, ast.FunctionDef) and class_node.name == function_name:
                            end_line = class_node.end_lineno if hasattr(class_node, 'end_lineno') else class_node.lineno + 5
                            return class_node.lineno, end_line
        except Exception as e:
            print(f"Error getting function line numbers: {e}")
            
        # If we reach here, the function wasn't found or there was an error
        return 0, 0
    
    def get_test_suites(self):
        """Get all test suites from the database."""
        self.cursor.execute("SELECT * FROM TestSuite ORDER BY creation_date DESC")
        return self.cursor.fetchall()
    
    def get_test_cases_by_function(self, function_name):
        """Get all test cases for a specific function."""
        self.cursor.execute(
            "SELECT tc.* FROM TestCase tc JOIN Function f ON tc.function_id = f.id WHERE f.name = ?",
            (function_name,)
        )
        return self.cursor.fetchall()
    
    def get_coverage_by_file(self, file_path):
        """Get coverage data for a specific file."""
        self.cursor.execute(
            "SELECT cd.* FROM CoverageData cd JOIN SourceFile sf ON cd.source_file_id = sf.id WHERE sf.path = ?", 
            (file_path,)
        )
        return self.cursor.fetchall()

    def get_test_file_data(self, file_path: str):
        """
        Retrieve all test cases, coverage data, and test results for a specific file.
        """
        query = """
            SELECT
                tc.id AS test_case_id,
                tc.name AS test_case_name,
                tc.test_function AS test_case_test_function,
                tc.test_method_type AS test_case_method_type,
                COALESCE(cd.missed_lines, 'None') AS coverage_data_missed_lines
            FROM SourceFile sf
            LEFT JOIN Function f ON sf.id = f.source_file_id
            LEFT JOIN TestCase tc ON f.id = tc.function_id
            LEFT JOIN TestResult tr ON tc.id = tr.test_case_id
            LEFT JOIN CoverageData cd ON sf.id = cd.source_file_id
            WHERE sf.path = ?;
        """
        self.cursor.execute(query, (file_path,))
        return self.cursor.fetchall()



import doctest
import importlib
import inspect
import json
import os
import re
import sqlite3
import sys
import time
import subprocess
import testgen.util.coverage_utils as coverage_utils
from types import ModuleType
from typing import List

import testgen
import testgen.util.file_utils as file_utils

from testgen.models.test_case import TestCase
from testgen.service.analysis_service import AnalysisService
from testgen.service.generator_service import GeneratorService
from testgen.sqlite.db_service import DBService
from testgen.models.analysis_context import AnalysisContext
from testgen.service.logging_service import get_logger


# Constants for test strategies
AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3
REINFORCE_STRAT = 4

# Constants for test formats
UNITTEST_FORMAT = 1
PYTEST_FORMAT = 2
DOCTEST_FORMAT = 3

class Service:
    def __init__(self):
        self.test_strategy: int = 0
        self.test_format: int = 0
        self.file_path = None
        self.generated_file_path = None
        self.class_name = None
        self.test_cases = []
        self.logger = get_logger()
        self.reinforcement_mode = "train"
        
        # Initialize specialized services
        self.analysis_service = AnalysisService()
        self.generator_service = GeneratorService(None, None, None)
        # Only initialize DB service if not running in Docker
        if os.environ.get("RUNNING_IN_DOCKER") is None:
            self.db_service = DBService()
        else:
            # Create a dummy DB service that doesn't do anything
            self.db_service = None

    def generate_test_cases(self) -> List[TestCase] | None:
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)

        if self.test_strategy == AST_STRAT:
            self.generated_file_path = self.generate_function_code()
            Service.wait_for_file(self.generated_file_path)
            self.analysis_service.set_file_path(self.generated_file_path)
            module = file_utils.load_module(self.generated_file_path)
            class_name = self.analysis_service.get_class_name(module)
            self.analysis_service.set_test_strategy(self.test_strategy, module.__name__, self.class_name)

        else:
            self.analysis_service.set_test_strategy(self.test_strategy, module.__name__, self.class_name)

        test_cases: List[TestCase] = []
        if self.test_strategy == REINFORCE_STRAT:
            test_cases = self.analysis_service.do_reinforcement_learning(self.file_path, class_name, self.reinforcement_mode)
        else:
            test_cases = self.analysis_service.generate_test_cases()

        if os.environ.get("RUNNING_IN_DOCKER") is not None:
            self.logger.debug(f"Serializing test cases {test_cases}")
            self.serialize_test_cases(test_cases)
            return None  # Exit early in analysis-only mode
        
        return test_cases

    def generate_tests(self, output_path=None):
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)

        test_cases = self.generate_test_cases()
        
        # Only process if we have test cases
        if test_cases is None:
            return None
            
        self.test_cases = test_cases

        # Only save to DB if not running in Docker
        if os.environ.get("RUNNING_IN_DOCKER") is None:
            file_path_to_use = self.generated_file_path if self.test_strategy == AST_STRAT else self.file_path
            self.db_service.save_test_generation_data(
                file_path_to_use,
                test_cases,
                self.test_strategy,
                class_name
            )

        test_file = self.generate_test_file(test_cases, output_path, module, class_name)

        # Ensure the test file is ready
        Service.wait_for_file(test_file)
        return test_file

    def generate_test_file(self, test_cases: List[TestCase], output_path: str | None = None,
                           module: ModuleType | None = None, class_name: str | None = None) -> str:
        if module is None:
            module = file_utils.load_module(self.file_path)
            class_name = self.analysis_service.get_class_name(module)

        if self.test_strategy == AST_STRAT:
            self.generator_service = GeneratorService(self.generated_file_path, output_path, self.test_format)
            self.generator_service.set_test_format(self.test_format)
            module = file_utils.load_module(self.generated_file_path)
        else:
            # Create the correct instance of the generator service
            self.generator_service = GeneratorService(self.file_path, output_path, self.test_format)
            self.generator_service.set_test_format(self.test_format)

        test_file = self.generator_service.generate_test_file(
            module,
            class_name,
            test_cases,
            output_path
        )

        print(f"Generated test file: {test_file}")

        # Ensure the test file is ready
        Service.wait_for_file(test_file)
        return test_file

    def generate_function_code(self):
        """Generate function code for a given class or module."""
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)
        functions = self.inspect_class(class_name)
        return self.generator_service.generate_function_code(self.file_path, class_name, functions)
    
    def run_coverage(self, test_file):
        """Run coverage analysis on the generated tests."""
        Service.wait_for_file(test_file)
        file_path_to_use = self.generated_file_path if self.test_strategy == AST_STRAT else self.file_path
        self.logger.debug(f"File path to use for coverage: {file_path_to_use}")
        coverage_output = ""
        
        try:
            if self.test_format == UNITTEST_FORMAT:
                self.execute_and_store_unittest(file_path_to_use, test_file)
            elif self.test_format == PYTEST_FORMAT:
                self.execute_and_store_pytest(test_file)
            elif self.test_format == DOCTEST_FORMAT:
                self.execute_and_store_doctest(test_file)
            else:
                raise ValueError("Unsupported test format for test results.")

            #Run coverage analysis
            subprocess.run(["python", "-m", "coverage", "run", "--source=.", test_file], check=True)
            result = subprocess.run(["python", "-m", "coverage", "report", file_path_to_use], check=True, capture_output=True, text=True)
            self._save_coverage_data(coverage_output, file_path_to_use)
            coverage_output = result.stdout

            self._save_coverage_data(coverage_output, file_path_to_use)

        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Error running coverage subprocess: {e}")

    def _save_coverage_data(self, coverage_output, file_path):
        """Parse coverage output and save to database."""
        # Skip if running in Docker or DB service is None
        if os.environ.get("RUNNING_IN_DOCKER") is not None or self.db_service is None:
            self.logger.debug("Skipping database operations in Docker container")
            return
        
        try:
            lines = coverage_output.strip().split('\n')
            if not lines:
                raise ValueError("No coverage data found in the output.")
            else:
                for line in lines:
                    if file_path in line:
                        parts = line.split()
                        if len(parts) >= 4:
                            file_name = os.path.basename(file_path)
                            try:
                                total_lines = int(parts[-3])
                                missed_lines = int(parts[-2])
                                executed_lines = total_lines - missed_lines
                                coverage_str = parts[-1].strip('%')
                                branch_coverage = float(coverage_str) / 100
                                
                                source_file_id = self._get_source_file_id(file_path)
                                
                                self.db_service.insert_coverage_data(
                                    file_name, 
                                    executed_lines, 
                                    missed_lines, 
                                    branch_coverage, 
                                    source_file_id
                                )
                                break
                            except (ValueError, IndexError) as e:
                                print(f"Error parsing coverage data: {e}")
        except Exception as e:
            print(f"Error saving coverage data: {e}")
            
    def _get_source_file_id(self, file_path):
        """Helper to get source file ID from DB."""
        conn = sqlite3.connect(self.db_service.db_name)
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        
        cursor.execute("SELECT id FROM SourceFile WHERE path = ?", (file_path,))
        row = cursor.fetchone()
        
        conn.close()
        
        if row:
            return row[0]
        else:
            with open(file_path, 'r') as f:
                lines_of_code = len(f.readlines())
            return self.db_service.insert_source_file(file_path, lines_of_code)

    def serialize_test_cases(self, test_cases):
        """Serialize input arguments ot JSON-compatible format"""
        print("##TEST_CASES_BEGIN##")
        serialized = []
        for tc in test_cases:
            case_data = {
                "func_name": tc.func_name,
                "inputs": self.serialize_value(tc.inputs),
                "expected": self.serialize_value(tc.expected)
            }
            serialized.append(case_data)
        print(json.dumps(serialized))
        print("##TEST_CASES_END##")

    def serialize_value(self, value):
        """Serialize a single value to a JSON-compatible format"""
        if value is None or isinstance(value, (bool, int, float, str)):
            return value
        elif isinstance(value, (list, tuple)):
            return [self.serialize_value(v) for v in value]
        else:
            return str(value)

    @staticmethod
    def parse_test_cases_from_logs(logs_output):
        """Extract and parse test cases from container logs"""
        pattern = r"##TEST_CASES_BEGIN##\n(.*?)\n##TEST_CASES_END##"
        match = re.search(pattern, logs_output, re.DOTALL)

        if not match:
            raise ValueError("Could not find test cases in the container logs")

        test_cases_json = match.group(1).strip()
        test_cases_data = json.loads(test_cases_json)

        test_cases: List[TestCase] = []
        for tc_data in test_cases_data:
            test_case = TestCase(
                func_name=tc_data["func_name"],
                inputs=tc_data["inputs"],
                expected=tc_data["expected"]
            )
            test_cases.append(test_case)

        return test_cases

    def set_file_path(self, path: str):
        """Set the file path for analysis and validate it."""
        print(f"Setting file path: {path}")
        if os.path.isfile(path) and path.endswith(".py"):
            self.file_path = path
            self.analysis_service.set_file_path(path)
        else:
            raise ValueError("Invalid file path! Please provide a valid Python file path.")

    def set_class_name(self, class_name: str):
        """Set the class name to analyze."""
        self.class_name = class_name

    def set_test_generator_format(self, test_format: int):
        """Set the test generator format."""
        self.test_format = test_format
        self.generator_service.set_test_format(test_format)

    def set_test_analysis_strategy(self, strategy: int):
        """Set the test analysis strategy."""
        self.test_strategy = strategy
        module = file_utils.load_module(self.file_path)
        self.analysis_service.set_test_strategy(strategy, module.__name__, self.class_name)
    
    def get_analysis_context(self, filepath: str) -> AnalysisContext:
        """Create an analysis context for the given file."""
        return self.analysis_service.create_analysis_context(filepath)

    @staticmethod
    def wait_for_file(file_path, retries=5, delay=1):
        """Wait for the generated file to appear."""
        while retries > 0 and not os.path.exists(file_path):
            time.sleep(delay)
            retries -= 1
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File '{file_path}' not found after waiting.")

    def get_full_import_path(self) -> str:
        """Get the full import path for the current file."""
        package_root = self.find_package_root()
        if not package_root:
            raise ImportError(f"Could not determine the package root for {self.file_path}.")

        module_path = os.path.abspath(self.file_path)
        rel_path = os.path.relpath(module_path, package_root)
        package_path = rel_path.replace(os.sep, ".")
        
        if package_path.endswith(".py"):
            package_path = package_path[:-3]

        package_name = os.path.basename(package_root)
        if not package_path.startswith(package_name + "."):
            package_path = package_name + "." + package_path

        return package_path

    def find_package_root(self):
        """Find the package root directory."""
        current_dir = os.path.abspath(os.path.dirname(self.file_path))
        last_valid = None

        while current_dir:
            if "__init__.py" in os.listdir(current_dir):
                last_valid = current_dir
            else:
                break

            parent_dir = os.path.dirname(current_dir)
            if parent_dir == current_dir:
                break
            current_dir = parent_dir

        return last_valid

    def inspect_class(self, class_name):
        """Inspect a class or module and return its functions."""
        module = file_utils.load_module(self.file_path)

        # Handle module-level functions when class_name is None
        if not class_name:
            # Get module-level functions
            functions = inspect.getmembers(module, inspect.isfunction)
            return functions

        # Handle class functions
        cls = getattr(module, class_name, None)
        if cls is None:
            raise ValueError(f"Class '{class_name}' not found in module '{module.__name__}'.")

        functions = inspect.getmembers(cls, inspect.isfunction)
        return functions

    def resolve_module_path(self, module_name):
        """Resolve a module name to its file path by checking multiple locations."""
        direct_path = f"/controller/{module_name}.py"
        if os.path.exists(direct_path):
            self.logger.debug(f"Found module at {direct_path}")
            return direct_path

        testgen_path = f"/controller/testgen/{module_name}.py"
        if os.path.exists(testgen_path):
            self.logger.debug(f"Found module at {testgen_path}")
            return testgen_path

        if '.' in module_name:
            parts = module_name.split('.')
            potential_path = os.path.join('/controller', *parts) + '.py'
            if os.path.exists(potential_path):
                self.logger.debug(f"Found module at {potential_path}")
                return potential_path

        self.logger.debug(f"Could not find module: {module_name}")
        return None

    def visualize_test_coverage(self):
        from testgen.service.cfg_service import CFGService
        cfg_service = CFGService()
        cfg_service.initialize_visualizer(self)

        return cfg_service.visualize_test_coverage(
            file_path=self.file_path,
            test_cases=self.test_cases,
        )

    def set_reinforcement_mode(self, mode: str):
        self.reinforcement_mode = mode
        if hasattr(self ,'analysis_service'):
            self.analysis_service.set_reinforcement_mode(mode)

    def _get_test_case_id(self, test_case_name: str) -> int:
        """
        Retrieve the test case ID from the database based on the test case name.
        Insert the test case if it does not exist.
        """
        if self.db_service is None:
            raise RuntimeError("Database service is not initialized.")

        # Query the database for the test case ID
        self.db_service.cursor.execute(
            "SELECT id FROM TestCase WHERE name = ?",
            (test_case_name,)
        )
        result = self.db_service.cursor.fetchone()

        if result:
            return result[0]  # Return the test case ID
        else:
            # Insert the test case into the database
            self.db_service.cursor.execute(
                "INSERT INTO TestCase (name) VALUES (?)",
                (test_case_name,)
            )
            self.db_service.conn.commit()
            return self.db_service.cursor.lastrowid

    def execute_and_store_pytest(self, test_file):
        import pytest
        from _pytest.reports import TestReport

        class PytestResultPlugin:
            def __init__(self, db_service, get_test_case_id):
                self.db_service = db_service
                self.get_test_case_id = get_test_case_id

            def pytest_runtest_logreport(self, report: TestReport):
                if report.when == "call":
                    test_case_id = self.get_test_case_id(report.nodeid)
                    status = report.outcome == "passed"
                    error_message = report.longreprtext if report.outcome == "failed" else None
                    self.db_service.insert_test_result(test_case_id, status, error_message)

        pytest.main([test_file], plugins=[PytestResultPlugin(self.db_service, self._get_test_case_id)])

        pytest.main([test_file])

    def execute_and_store_unittest(self, file_path_to_use, test_file):
        import unittest
        loader = unittest.TestLoader()
        self.logger.debug(f"Discovering tests in: {os.path.dirname(file_path_to_use)} with pattern: {os.path.basename(test_file)}")
        test_module = os.path.relpath(test_file,
                                      start=os.getcwd())  # Get relative path from the current working directory
        test_module = test_module.replace("/", ".").replace("\\", ".").rstrip(".py")  # Convert to module name
        if test_module.startswith("."):
            test_module = test_module[1:]  # Remove leading dot if present
        self.logger.debug(f"Test module: {test_module}")
        suite = loader.loadTestsFromName(test_module)
        runner = unittest.TextTestRunner()
        result = runner.run(suite)

        for test_case, traceback in result.failures + result.errors:
            test_case_id = self._get_test_case_id(str(test_case))
            self.db_service.insert_test_result(test_case_id, status=False, error=traceback)

        successful_tests = set(str(test) for test in suite) - set(
            str(test) for test, _ in result.failures + result.errors)
        for test_case in successful_tests:
            test_case_id = self._get_test_case_id(str(test_case))
            self.db_service.insert_test_result(test_case_id, status=True, error=None)

    def execute_and_store_doctest(self, test_file):
        module_name = os.path.splitext(os.path.basename(test_file))[0]
        spec = importlib.util.spec_from_file_location(module_name, test_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Now run doctests on the loaded module
        result = doctest.testmod(module)

        test_case_id = self._get_test_case_id(test_file)
        status = result.failed == 0
        error_message = f"{result.failed} of {result.attempted} tests failed" if result.failed > 0 else None
        self.db_service.insert_test_result(test_case_id, status, error_message)

    def query_test_file_data(self, test_file_name: str):
        if self.db_service is None:
            raise RuntimeError("Database service is not initialized.")

        results = self.db_service.get_test_file_data(test_file_name)
        if not results:
            self.logger.debug(f"No data found for file: {test_file_name}")
            return

        from tabulate import tabulate
        rows = [dict(row) for row in results]

        print(f"Results for file: {test_file_name}")
        print(tabulate(rows, headers="keys", tablefmt="grid"))

    def select_all_from_db(self) -> None:
        rows = self.db_service.get_test_suites()
        for row in rows:
            print(repr(dict(row)))



from typing import List, Dict
import inspect
import ast
import astor
from testgen.tree.node import Node

class CodeGenerator:
    def __init__(self):
        pass
        
    def is_all_boolean_params(self, param_types: Dict[str, str]) -> bool:
        """Check if all parameters are boolean type."""
        if not param_types:
            return False
            
        for param_type in param_types.values():
            if param_type != 'bool':
                return False
        return True

    def get_original_function_source(self, func_obj) -> str:
        """Extract the original source code of the function."""
        try:
            # Get the source code as a string
            source = inspect.getsource(func_obj)
            return source
        except (IOError, TypeError) as e:
            print(f"Error getting source: {e}")
            return None

    def generate_code_from_tree(self, func_name: str, root: Node, params: List[str], operation, is_class_method: bool, param_types: Dict[str, str] = None) -> str:
        # If we have param types and not all are boolean, extract the original source
        if param_types and not self.is_all_boolean_params(param_types):
            original_source = self.get_original_function_source(operation)
            if original_source:
                return original_source
                
        # The existing tree-based generation for boolean params
        def traverse(node, depth, path=[], indent_level=1):
            base_indent = "    " * indent_level

            if depth == len(params):
                path = path + [node.value]
                if is_class_method:
                    result = operation(self, *path)
                else:
                    result = operation(*path)
                if isinstance(result, str):
                    return f"{base_indent}return '{result}'\n"
                else:
                    return f"{base_indent}return {result}\n"

            param = params[depth]

            if isinstance(node.value, bool):
                path = path + [node.value]

            true_branch = f"{base_indent}if {param} == True:\n"
            false_branch = f"{base_indent}else:\n"

            true_code = traverse(node.children[0], depth + 1, path, indent_level + 1)
            false_code = traverse(node.children[1], depth + 1, path, indent_level + 1)

            return f"{true_branch}{true_code}{false_branch}{false_code}"

        typed_param_list = []
        if is_class_method:
            typed_param_list.append("self")

        for param in params:
            param_type = param_types.get(param, 'bool') if param_types else 'bool'
            typed_param_list.append(f"{param}: {param_type}")

        function_code = f"def {func_name}({', '.join(typed_param_list)}):\n"
        body_code = traverse(root, 0)

        return function_code + body_code

    def generate_class(self, class_name):
        branched_class_name = f"Generated{class_name}"
        file_path = f"generated_{class_name.lower()}.py"
        class_file = open(f"{file_path}", "w")
        class_file.write(f"class {branched_class_name}:\n")
        return class_file


import ast
from types import ModuleType


class FunctionMetadata:
    def __init__(self, filename: str, module: ModuleType, class_name: str, function_name: str, func_def: ast.FunctionDef, params: dict):
        self._filename: str = filename
        self._module: ModuleType = module
        self._class_name: str = class_name
        self._function_name: str = function_name
        self._func_def: ast.FunctionDef = func_def
        self._params: dict = params
    
    @property
    def filename(self) -> str:
        return self._filename
    
    @filename.setter
    def filename(self, filename: str):
        self._filename = filename
    
    @property
    def module(self) -> ModuleType:
        return self._module
    
    @module.setter
    def module(self, module: ModuleType):
        self._module = module

    @property
    def class_name(self) -> str:
        return self._class_name
    
    @class_name.setter
    def class_name(self, class_name: str):
        self._class_name = class_name
    
    @property
    def function_name(self) -> str:
        return self._function_name
    
    @function_name.setter
    def function_name(self, func_name: str):
        self._function_name = func_name
    
    @property
    def func_def(self) -> ast.FunctionDef:
        return self._func_def
    
    @func_def.setter
    def func_def(self, func_def: ast.FunctionDef):
        self._func_def = func_def
    
    @property
    def params(self) -> dict:
        return self._params
    
    @params.setter
    def params(self, params: dict):
        self._params = params




import inspect
import ast
import time
from types import ModuleType
from typing import Dict, List

import testgen
from testgen.service.logging_service import get_logger
import testgen.util.file_utils
import testgen.util.file_utils as file_utils
import testgen.util.utils
from testgen.analyzer.ast_analyzer import ASTAnalyzer
from testgen.analyzer.fuzz_analyzer import FuzzAnalyzer
from testgen.analyzer.random_feedback_analyzer import RandomFeedbackAnalyzer
from testgen.models.test_case import TestCase
from testgen.analyzer.test_case_analyzer_context import TestCaseAnalyzerContext
from testgen.reinforcement.agent import ReinforcementAgent
from testgen.reinforcement.environment import ReinforcementEnvironment
from testgen.reinforcement.statement_coverage_state import StatementCoverageState
from testgen.models.analysis_context import AnalysisContext
from testgen.models.function_metadata import FunctionMetadata

# Constants for test strategies
AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3
REINFORCE_STRAT = 4

class AnalysisService:
    def __init__(self):
        self.file_path = None
        self.class_name = None
        self.test_case_analyzer_context = TestCaseAnalyzerContext(None, None)
        self.test_strategy = 0
        self.reinforcement_mode = "train"
        self.logger = get_logger()

    def generate_test_cases(self) -> List[TestCase]:
            """Generate test cases using the current strategy."""
            if self.test_strategy == REINFORCE_STRAT:
                return self.do_reinforcement_learning(self.file_path)
            else:
                self.test_case_analyzer_context.do_logic()
                return self.test_case_analyzer_context.test_cases
    
    def create_analysis_context(self, filepath: str) -> AnalysisContext:
        """Create an analysis context for the given file."""
        self.logger.debug(f"Creating analysis context for {filepath}")
        filename = file_utils.get_filename(filepath)
        self.logger.debug(f"Filename: {filename}")
        module = file_utils.load_module(filepath)
        self.logger.debug(f"Module: {module}")
        class_name = self.get_class_name(module)
        self.logger.debug(f"Class name: {class_name}")
        function_data = self.get_function_data(filename, module, class_name)
        self.logger.debug(f"Function data: {function_data}")
        return AnalysisContext(filepath, filename, class_name, module, function_data)
    
    def get_function_data(self, filename: str, module: ModuleType, class_name: str | None) -> List[FunctionMetadata]:
        function_metadata_list: List[FunctionMetadata] = []

        # Parse the module's source code
        source_code = inspect.getsource(module)
        tree = ast.parse(source_code)

        # Find all function definitions in the module
        for node in tree.body:
            if isinstance(node, ast.FunctionDef):
                function_metadata_list.append(self._create_function_metadata(filename, module, class_name, node))
            elif isinstance(node, ast.ClassDef):
                # Get the class name to check if it matches the expected class
                ast_class_name = node.name

                # If we have a specified class and it doesn't match, skip
                if self.class_name and ast_class_name != self.class_name:
                    continue

                # Process class methods
                for class_node in node.body:
                    if isinstance(class_node, ast.FunctionDef):
                        # Skip private methods (those starting with _)
                        if not class_node.name.startswith('_'):
                            function_metadata_list.append(self._create_function_metadata(filename, module, class_name, class_node))

        return function_metadata_list

    def do_reinforcement_learning(self, filepath: str, class_name: str | None, mode: str = None) -> List[TestCase]:
        mode = mode or self.reinforcement_mode
        module: ModuleType = testgen.util.file_utils.load_module(filepath)
        tree: ast.Module = testgen.util.file_utils.load_and_parse_file_for_tree(filepath)
        functions: List[ast.FunctionDef] = testgen.util.utils.get_functions(tree)
        self.class_name = class_name
        time_limit: int = 30
        all_test_cases: List[TestCase] = []

        q_table = self._load_q_table()

        for function in functions:
            print(f"\nStarting reinforcement learning for function {function.name}")
            start_time = time.time()
            function_test_cases: List[TestCase] = []
            best_coverage: float = 0.0

            # Create environment and agent once per function
            environment = ReinforcementEnvironment(filepath, function, module, self.class_name, function_test_cases, state=StatementCoverageState(None))
            environment.state = StatementCoverageState(environment)

            # Create agent with existing Q-table
            agent = ReinforcementAgent(filepath, environment, function_test_cases, q_table)

            if mode == "train":
                new_test_cases = agent.do_q_learning()
                function_test_cases.extend(new_test_cases)
            else:
                new_test_cases = agent.collect_test_cases()
                function_test_cases.extend(new_test_cases)

            print(f"\nNumber of test cases for {function.name}: {len(function_test_cases)}")

            current_coverage: float = environment.run_tests()
            print(f"Current coverage: {function.name}: {current_coverage}")

            q_table.update(agent.q_table)

            # Process and filter unique test cases
            seen = set()
            unique_test_cases: List[TestCase] = []
            for case in function_test_cases:
                # Make case tuple hashable
                case_inputs = tuple(case.inputs) if isinstance(case.inputs, list) else case.inputs
                case_key = (case.func_name, case_inputs)
                if case_key not in seen:
                    seen.add(case_key)
                    unique_test_cases.append(case)

            all_test_cases.extend(unique_test_cases)
            print(f"Final coverage for {function.name}: {best_coverage}%")
            print(f"Final test cases for {function.name}: {len(unique_test_cases)}")

        self._save_q_table(q_table)

        print("\nReinforcement Learning Complete")
        print(f"Total test cases found: {len(all_test_cases)}")
        return all_test_cases

    def _create_function_metadata(self, filename: str, module: ModuleType, class_name: str | None,
                                  func_node: ast.FunctionDef) -> FunctionMetadata:
        function_name = func_node.name

        param_types = self._get_params(func_node)

        return FunctionMetadata(filename, module, class_name, function_name, func_node, param_types)
    
    def _get_params(self, func_node: ast.FunctionDef) -> Dict[str, str]:
        # Extract parameter types
        param_types = {}
        for arg in func_node.args.args:
            param_name = arg.arg
            if param_name == 'self':
                continue

            if arg.annotation:
                param_type = ast.unparse(arg.annotation)
                param_types[param_name] = param_type
            else:
                param_types[param_name] = None
        return param_types
    
    def set_test_strategy(self, strategy: int, module_name: str, class_name: str):
        """Set the test analysis strategy."""
        self.test_strategy = strategy
        analysis_context = self.create_analysis_context(self.file_path)

        if strategy == AST_STRAT:
            analyzer = ASTAnalyzer(analysis_context)
            self.test_case_analyzer_context = TestCaseAnalyzerContext(analysis_context, analyzer)
        elif strategy == FUZZ_STRAT:
            analyzer = FuzzAnalyzer(analysis_context)
            self.test_case_analyzer_context = TestCaseAnalyzerContext(analysis_context, analyzer)
        elif strategy == RANDOM_STRAT:
            analyzer = RandomFeedbackAnalyzer(analysis_context)
            self.test_case_analyzer_context = TestCaseAnalyzerContext(analysis_context, analyzer)
        elif strategy == REINFORCE_STRAT:
            pass
        else:
            raise NotImplementedError(f"Test strategy {strategy} not implemented")
        
    def set_file_path(self, path: str):
        """Set the file path for analysis."""
        self.file_path = path

    def set_reinforcement_mode(self, mode: str):
        self.reinforcement_mode = mode

    @staticmethod
    def get_class_name(module: ModuleType) -> str | None:
        """Get class from module otherwise return None."""
        for name, cls in inspect.getmembers(module, inspect.isclass):
            if cls.__module__ == module.__name__:
                return name
        return None

    @staticmethod
    def _save_q_table(q_table):
        """Save Q-table to a global JSON file"""
        import json
        import os

        q_table_dir = "q_table"
        os.makedirs(q_table_dir, exist_ok=True)
        q_table_path = os.path.join(q_table_dir, "global_q_table.json")

        global_q_table = {}
        if os.path.exists(q_table_path):
            try:
                with open(q_table_path, 'r') as f:
                    global_q_table = json.load(f)
            except Exception as e:
                print(f"Error loading existing Q-table: {e}")

        serializable_q_table = {}
        for key, value in q_table.items():
            state, action = key
            state_str = str(state)
            serializable_q_table[f"{state_str}|{action}"] = value

        global_q_table = serializable_q_table  # Replace with latest Q-table

        try:
            with open(q_table_path, 'w') as f:
                json.dump(global_q_table, f)
            print(f"Q-table saved to {q_table_path}")
        except Exception as e:
            print(f"Error saving Q-table: {e}")

    @staticmethod
    def _load_q_table():
        """Load Q-table from the global JSON file"""
        import json
        import os
        import ast

        q_table_dir = "q_table"
        q_table_path = os.path.join(q_table_dir, "global_q_table.json")

        if not os.path.exists(q_table_path):
            print(f"No existing Q-table found at {q_table_path}")
            return {}

        try:
            with open(q_table_path, 'r') as f:
                serialized_q_table = json.load(f)

            # Convert serialized keys back to (state, action) tuples
            q_table = {}
            for key, value in serialized_q_table.items():
                state_str, action = key.split('|')
                try:
                    state = ast.literal_eval(state_str)
                    q_table[(state, action)] = value
                except (ValueError, SyntaxError):
                    print(f"Skipping invalid state: {state_str}")

            print(f"Loaded global Q-table with {len(q_table)} entries")
            return q_table
        except Exception as e:
            print(f"Error loading Q-table: {e}")
            return {}


class GeneratedSampleCodeBin:
    def bin_and(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_and(True, True)
        True

        >>> GeneratedSampleCodeBin().bin_and(True, False)
        False

        >>> GeneratedSampleCodeBin().bin_and(False, True)
        False
        """
        if a == True:
            if b == True:
                return True
            else:
                return False
        else:
            if b == True:
                return False
            else:
                return False
    

    def bin_and_bad(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_and_bad(True, True)
        True

        >>> GeneratedSampleCodeBin().bin_and_bad(True, False)
        True

        >>> GeneratedSampleCodeBin().bin_and_bad(False, True)
        True
        """
        if a == True:
            if b == True:
                return True
            else:
                return True
        else:
            if b == True:
                return True
            else:
                return False
    

    def bin_nand(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_nand(True, True)
        False

        >>> GeneratedSampleCodeBin().bin_nand(True, False)
        True

        >>> GeneratedSampleCodeBin().bin_nand(False, True)
        True
        """
        if a == True:
            if b == True:
                return False
            else:
                return True
        else:
            if b == True:
                return True
            else:
                return True
    

    def bin_nor(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_nor(True, True)
        False

        >>> GeneratedSampleCodeBin().bin_nor(True, False)
        False

        >>> GeneratedSampleCodeBin().bin_nor(False, True)
        False
        """
        if a == True:
            if b == True:
                return False
            else:
                return False
        else:
            if b == True:
                return False
            else:
                return True
    

    def bin_or(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_or(True, True)
        True

        >>> GeneratedSampleCodeBin().bin_or(True, False)
        True

        >>> GeneratedSampleCodeBin().bin_or(False, True)
        True
        """
        if a == True:
            if b == True:
                return True
            else:
                return True
        else:
            if b == True:
                return True
            else:
                return False
    

    def bin_or_bad(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_or_bad(True, True)
        False

        >>> GeneratedSampleCodeBin().bin_or_bad(True, False)
        False

        >>> GeneratedSampleCodeBin().bin_or_bad(False, True)
        False
        """
        if a == True:
            if b == True:
                return False
            else:
                return False
        else:
            if b == True:
                return False
            else:
                return True
    

    def bin_xor(self, a: bool, b: bool):
        """
        >>> GeneratedSampleCodeBin().bin_xor(True, True)
        False

        >>> GeneratedSampleCodeBin().bin_xor(True, False)
        True

        >>> GeneratedSampleCodeBin().bin_xor(False, True)
        True
        """
        if a == True:
            if b == True:
                return False
            else:
                return True
        else:
            if b == True:
                return True
            else:
                return False
    

    def mux(self, c1: bool, c2: bool, x0: bool, x1: bool, x2: bool, x3: bool):
        """
        >>> GeneratedSampleCodeBin().mux(True, True, True, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, True, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, True, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, True, False, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, False, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, False, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, False, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, True, False, False, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, True, False, True, True, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, True, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, True, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, True, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, False, True, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, False, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, False, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, True, False, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, True, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, True, True, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, True, True, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, True, True, False, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, True, False, True, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, True, False, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, True, False, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, True, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, False, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, False, True, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, False, True, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, False, True, False, False)
        True

        >>> GeneratedSampleCodeBin().mux(True, False, False, False, True, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, False, False, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, False, False, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(True, False, False, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, True, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, True, True, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, True, True, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, True, True, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, True, False, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, True, False, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, True, False, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, True, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, False, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, False, True, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, False, True, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, False, True, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, False, False, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, False, False, True, False)
        True

        >>> GeneratedSampleCodeBin().mux(False, True, False, False, False, True)
        False

        >>> GeneratedSampleCodeBin().mux(False, True, False, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, True, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, True, True, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, True, True, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, True, True, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, True, False, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, True, False, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, True, False, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, True, False, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, False, True, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, False, True, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, False, True, False, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, False, True, False, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, False, False, True, True)
        True

        >>> GeneratedSampleCodeBin().mux(False, False, False, False, True, False)
        False

        >>> GeneratedSampleCodeBin().mux(False, False, False, False, False, True)
        True
        """
        if c1 == True:
            if c2 == True:
                if x0 == True:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return True
                            else:
                                return True
                    else:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return True
                            else:
                                return True
                else:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return False
                            else:
                                return False
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
                    else:
                        if x2 == True:
                            if x3 == True:
                                return False
                            else:
                                return False
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
            else:
                if x0 == True:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return True
                            else:
                                return True
                    else:
                        if x2 == True:
                            if x3 == True:
                                return False
                            else:
                                return False
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
                else:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return True
                            else:
                                return True
                    else:
                        if x2 == True:
                            if x3 == True:
                                return False
                            else:
                                return False
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
        else:
            if c2 == True:
                if x0 == True:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
                    else:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
                else:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
                    else:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return True
                        else:
                            if x3 == True:
                                return False
                            else:
                                return False
            else:
                if x0 == True:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return False
                        else:
                            if x3 == True:
                                return True
                            else:
                                return False
                    else:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return False
                        else:
                            if x3 == True:
                                return True
                            else:
                                return False
                else:
                    if x1 == True:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return False
                        else:
                            if x3 == True:
                                return True
                            else:
                                return False
                    else:
                        if x2 == True:
                            if x3 == True:
                                return True
                            else:
                                return False
                        else:
                            if x3 == True:
                                return True
                            else:
                                return False
    




import ast
import importlib
import random
import time
import traceback
from typing import List, Dict, Set
import z3

import testgen.util.randomizer
import testgen.util.utils as utils
import testgen.util.coverage_utils as coverage_utils
from testgen.analyzer.contracts.contract import Contract
from testgen.analyzer.contracts.no_exception_contract import NoExceptionContract
from testgen.analyzer.contracts.nonnull_contract import NonNullContract
from testgen.models.test_case import TestCase
from testgen.analyzer.test_case_analyzer import TestCaseAnalyzerStrategy
from abc import ABC

from testgen.models.function_metadata import FunctionMetadata
from testgen.util.z3_utils.constraint_extractor import extract_branch_conditions
from testgen.util.z3_utils.ast_to_z3 import ast_to_z3_constraint


# Citation in which this method and algorithm were taken from:
# C. Pacheco, S. K. Lahiri, M. D. Ernst and T. Ball, "Feedback-Directed Random Test Generation," 29th International
# Conference on Software Engineering (ICSE'07), Minneapolis, MN, USA, 2007, pp. 75-84, doi: 10.1109/ICSE.2007.37.
# keywords: {System testing;Contracts;Object oriented modeling;Law;Legal factors;Open source software;Software
# testing;Feedback;Filters;Error correction codes},

class RandomFeedbackAnalyzer(TestCaseAnalyzerStrategy, ABC):
    def __init__(self, analysis_context=None):
        super().__init__(analysis_context)
        self.test_cases = []
        self.covered_lines: Dict[str, Set[int]] = {}
        self.covered_functions: Set[str] = set()

    def collect_test_cases(self, function_metadata: FunctionMetadata, time_limit: int = 5) -> List[TestCase]:
        self.test_cases = []
        start_time = time.time()
        
        while (time.time() - start_time) < time_limit:
            
            try:
                param_values = self.generate_random_inputs(function_metadata.params)
                module = self.analysis_context.module
                func_name = function_metadata.function_name
                
                if self._analysis_context.class_name:
                    cls = getattr(module, self._analysis_context.class_name)
                    obj = cls()
                    function = getattr(obj, func_name)
                else:
                    function = getattr(module, func_name)
                
                import inspect
                sig = inspect.signature(function)
                param_names = [p.name for p in sig.parameters.values() if p.name != 'self']
                
                ordered_args = []
                for name in param_names:
                    if name in param_values:
                        ordered_args.append(param_values[name])
                    else:
                        ordered_args.append(None)
                
                result = function(*ordered_args)
                test_case = TestCase(func_name, tuple(ordered_args), result)

                if not self.is_duplicate_test_case(test_case):
                    self.test_cases.append(test_case)
                
                    covered = self.covered(function_metadata)
                    if covered:
                        break
                else:
                    # Optionally log duplicate detection
                    self.logger.debug(f"Skipping duplicate test case: {func_name}{test_case.inputs}")
                            
            except Exception as e:
                print(f"Error testing {function_metadata.function_name}: {e}")

        return self.test_cases
    
    def is_duplicate_test_case(self, new_test_case: TestCase) -> bool:
        for existing_test_case in self.test_cases:
            if (existing_test_case.func_name == new_test_case.func_name and
                existing_test_case.inputs == new_test_case.inputs):
                return True
        return False

    def covered(self, func: FunctionMetadata) -> bool:
        if func.function_name not in self.covered_lines:
            self.covered_lines[func.function_name] = set()

        for test_case in [tc for tc in self.test_cases if tc.func_name == func.function_name]:
            analysis = coverage_utils.get_coverage_analysis(self._analysis_context.filepath, self._analysis_context.class_name,
                                                                         func.function_name, test_case.inputs)
            covered = coverage_utils.get_list_of_covered_statements(analysis)
            self.covered_lines[func.function_name].update(covered)
            self.logger.debug(f"Covered lines for {func.function_name}: {self.covered_lines[func.function_name]}")

        executable_statements = set(self.get_all_executable_statements(func))
        self.logger.debug(f"Executable statements for {func.function_name}: {executable_statements}")

        return self.covered_lines[func.function_name] == executable_statements

    def execute_sequence(self, sequence, contracts: List[Contract]):
        """Execute a sequence and check contract violations"""
        func_name, args_dict = sequence
        args = tuple(args_dict.values())  # Convert dict values to tuple

        try:
            # Use module from analysis context if available
            module = self.analysis_context.module

            if self._analysis_context.class_name:
                cls = getattr(module, self._analysis_context.class_name, None)
                if cls is None:
                    raise AttributeError(f"Class '{self._analysis_context.class_name}' not found")
                obj = cls()  # Instantiate the class
                func = getattr(obj, func_name, None)

                import inspect
                sig = inspect.signature(func)
                param_names = [p.name for p in sig.parameters.values() if p.name != 'self']
            else:
                func = getattr(module, func_name, None)

                import inspect
                sig = inspect.signature(func)
                param_names = [p.name for p in sig.parameters.values()]

            # Create ordered arguments based on function signature
            ordered_args = []
            for name in param_names:
                if name in args_dict:
                    ordered_args.append(args_dict[name])

            # Check preconditions
            for contract in contracts:
                if not contract.check_preconditions(tuple(ordered_args)):
                    print(f"Preconditions failed for {func_name} with {tuple(ordered_args)}")
                    return None, True

            # Execute function with properly ordered arguments
            output = func(*ordered_args)
            exception = None

        except Exception as e:
            print(f"EXCEPTION IN RANDOM FEEDBACK: {e}")
            print(traceback.format_exc())
            output = None
            exception = e

        # Check postconditions
        for contract in contracts:
            if not contract.check_postconditions(tuple(ordered_args), output, exception):
                print(f"Postcondition failed for {func_name} with {tuple(ordered_args)}")
                return output, True

        return output, False
    

    # TODO: Currently only getting random vals of primitives, extend to sequences
    def random_seqs_and_vals(self, param_types, non_error_seqs=None):
        return self.generate_random_inputs(param_types)

    @staticmethod
    def extract_parameter_types(func_node):
        """Extract parameter types from a function node."""
        param_types = {}
        for arg in func_node.args.args:
            param_name = arg.arg
            if arg.annotation:
                param_type = ast.unparse(arg.annotation)
                param_types[param_name] = param_type
            else:
                if param_name != 'self':
                    param_types[param_name] = None
        return param_types

    @staticmethod
    def generate_random_inputs(param_types):
        """Generate inputs for fuzzing based on parameter types."""
        inputs = {}
        for param, param_type in param_types.items():
            if param_type == "int":
                random_integer = random.randint(-500, 500)  # Wider range for better edge cases
                inputs[param] = random_integer
            elif param_type == "bool":
                random_choice = random.choice([True, False])
                inputs[param] = random_choice
            elif param_type == "float":
                random_float = random.uniform(-500.0, 500.0)  # Wider range for better edge cases
                inputs[param] = random_float
            elif param_type == "str":
                # Generate diverse strings instead of always "abc"
                string_type = random.choice([
                    "empty", "short", "medium", "long", "special", "numeric", "whitespace"
                ])
                
                if string_type == "empty":
                    inputs[param] = ""
                elif string_type == "short":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyz', k=random.randint(1, 3)))
                elif string_type == "medium":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', k=random.randint(4, 10)))
                elif string_type == "long":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789', k=random.randint(11, 30)))
                elif string_type == "special":
                    inputs[param] = ''.join(random.choices('!@#$%^&*()_+-=[]{}|;:,./<>?', k=random.randint(1, 8)))
                elif string_type == "numeric":
                    inputs[param] = ''.join(random.choices('0123456789', k=random.randint(1, 10)))
                else:  # whitespace
                    inputs[param] = ' ' * random.randint(1, 5)
            else:
                # For unknown types, try a default value
                inputs[param] = None
                
        return inputs
    
    # Algorithm described in above article
    # Classes is the classes for which we want to generate sequences
    # Contracts express invariant properties that hold both at entry and exit from a call
        # Contract takes as input the current state of the system (runtime values created in the sequence so far, and any exception thrown by the last call), and returns satisfied or violated
        # Output is the runtime values and boolean flag violated
    # Filters determine which values of a sequence are extensible and should be used as inputs
    def generate_sequences(self, function_metadata: List[FunctionMetadata], classes=None, contracts: List[Contract] = None, filters=None, time_limit=20):
        contracts = [NonNullContract(), NoExceptionContract()]
        error_seqs = [] # execution violates a contract
        non_error_seqs = [] # execution does not violate a contract

        functions = self._analysis_context.function_data
        start_time = time.time()
        while(time.time() - start_time) >= time_limit:
            # Get random function
            func = random.choice(functions)
            param_types: dict = func.params
            vals: dict = self.random_seqs_and_vals(param_types)
            new_seq = (func.function_name, vals)
            if new_seq in error_seqs or new_seq in non_error_seqs:
                continue
            outs_violated: tuple = self.execute_sequence(new_seq, contracts)
            violated: bool = outs_violated[1]
            # Create tuple of sequence ((func name, args), output)
            new_seq_out = (new_seq, outs_violated[0])
            if violated:
                error_seqs.append(new_seq_out)
            else:
                # Question: Should I use the failed contract to be the assertion in unit test??
                non_error_seqs.append(new_seq_out)
        return error_seqs, non_error_seqs

    def generate_sequences_new(self, contracts: List[Contract] = None, filters=None, time_limit=20):
        contracts = [NonNullContract(), NoExceptionContract()]
        error_seqs = []  # execution violates a contract
        non_error_seqs = []  # execution does not violate a contract

        functions = self._analysis_context.function_data.copy()
        start_time = time.time()

        while (time.time() - start_time) < time_limit:
            # Get random function
            func = random.choice(functions)
            param_types: dict = func.params
            vals: dict = self.random_seqs_and_vals(param_types)
            new_seq = (func.function_name, vals)

            if new_seq in [seq[0] for seq in error_seqs] or new_seq in [seq[0] for seq in non_error_seqs]:
                continue

            outs_violated: tuple = self.execute_sequence(new_seq, contracts)
            violated: bool = outs_violated[1]

            # Create tuple of sequence ((func name, args), output)
            new_seq_out = (new_seq, outs_violated[0])

            if violated:
                error_seqs.append(new_seq_out)

            else:
                non_error_seqs.append(new_seq_out)

            test_case = TestCase(new_seq_out[0][0], tuple(new_seq_out[0][1].values()), new_seq_out[1])
            self.test_cases.append(test_case)
            fully_covered = self.covered(func)
            if fully_covered:
                print(f"Function {func.function_name} is fully covered")
                functions.remove(func)

            if not functions:
                self.test_cases.sort(key=lambda tc: tc.func_name)
                print("All functions covered")
                break

        self.test_cases.sort(key=lambda tc: tc.func_name)
        return error_seqs, non_error_seqs
    
    def get_all_executable_statements(self, func: FunctionMetadata):
        import ast

        test_cases = [tc for tc in self.test_cases if tc.func_name == func.function_name]

        if not test_cases:
            print("Warning: No test cases available to determine executable statements")
            from testgen.util.randomizer import new_random_test_case
            temp_case = new_random_test_case(self._analysis_context.filepath, func.func_def)
            analysis = coverage_utils.get_coverage_analysis(self._analysis_context.filepath, self._analysis_context.class_name, func.function_name,
                                                                         temp_case.inputs)
        else:
            analysis = coverage_utils.get_coverage_analysis(self._analysis_context.filepath, self._analysis_context.class_name, func.function_name, test_cases[0].inputs)

        executable_lines = list(analysis[1])

        with open(self._analysis_context.filepath, 'r') as f:
            source = f.read()

        tree = ast.parse(source)

        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef) and node.name == func.func_def.name:
                for if_node in ast.walk(node):
                    if isinstance(if_node, ast.If) and if_node.orelse:
                        if isinstance(if_node.orelse[0], ast.If):
                            continue
                        else_line = if_node.orelse[0].lineno - 1

                        with open(self._analysis_context.filepath, 'r') as f:
                            lines = f.readlines()
                            if else_line <= len(lines):
                                line_content = lines[else_line - 1].strip()
                                if line_content == "else:":
                                    if else_line not in executable_lines:
                                        executable_lines.append(else_line)

        return sorted(executable_lines)

    """
    def collect_test_cases_with_z3(self, function_metadata: FunctionMetadata) -> List[TestCase]:
        test_cases = []
        
        z3_test_cases = self.generate_z3_test_cases(function_metadata)
        if z3_test_cases:
            test_cases.extend(z3_test_cases)
        
        if not test_cases:
            test_cases = self.generate_sequences_new()[1]
        
        self.test_cases = test_cases
        return test_cases

    def generate_z3_test_cases(self, function_metadata: FunctionMetadata) -> List[TestCase]:
        test_cases = []
        
        branch_conditions, param_types = extract_branch_conditions(function_metadata.func_def)
        
        if not branch_conditions:
            random_inputs = self.generate_random_inputs(function_metadata.params)
            try:
                module = self.analysis_context.module
                func_name = function_metadata.function_name
                
                if self._analysis_context.class_name:
                    cls = getattr(module, self._analysis_context.class_name)
                    obj = cls()
                    func = getattr(obj, func_name)
                    ordered_args = self._order_arguments(func, random_inputs)
                    output = func(*ordered_args)
                else:
                    func = getattr(module, func_name)
                    ordered_args = self._order_arguments(func, random_inputs)
                    output = func(*ordered_args)
                
                test_cases.append(TestCase(func_name, tuple(ordered_args), output))
            except Exception as e:
                print(f"Error executing function with random inputs: {e}")
            
            return test_cases
        
        for branch_condition in branch_conditions:
            try:
                z3_expr, z3_vars = ast_to_z3_constraint(branch_condition, function_metadata.params)
                
                solver = z3.Solver()
                solver.add(z3_expr)
                
                neg_solver = z3.Solver()
                neg_solver.add(z3.Not(z3_expr))
                
                for current_solver in [solver, neg_solver]:
                    if current_solver.check() == z3.sat:
                        model = current_solver.model()
                        
                        param_values = self._extract_z3_solution(model, z3_vars, function_metadata.params)
                        
                        ordered_params = self._order_parameters(function_metadata.func_def, param_values)
                        
                        try:
                            module = self.analysis_context.module
                            func_name = function_metadata.function_name
                            
                            if self._analysis_context.class_name:
                                cls = getattr(module, self._analysis_context.class_name)
                                obj = cls()
                                func = getattr(obj, func_name)
                            else:
                                func = getattr(module, func_name)
                                
                            result = func(*ordered_params)
                            test_cases.append(TestCase(func_name, tuple(ordered_params), result))
                        except Exception as e:
                            print(f"Error executing function with Z3 solution: {e}")
                            self._add_random_test_case(function_metadata, test_cases)
                    else:
                        self._add_random_test_case(function_metadata, test_cases)
                        
            except Exception as e:
                print(f"Error processing branch condition with Z3: {e}")
                self._add_random_test_case(function_metadata, test_cases)
        
        return test_cases
    
    def _extract_z3_solution(self, model, z3_vars, param_types):
        param_values = {}
        
        for var_name, z3_var in z3_vars.items():
            if var_name in param_types:
                try:
                    model_value = model.evaluate(z3_var)
                    
                    if param_types[var_name] == "int":
                        param_values[var_name] = model_value.as_long()
                    elif param_types[var_name] == "float":
                        param_values[var_name] = float(model_value.as_decimal(10))
                    elif param_types[var_name] == "bool":
                        param_values[var_name] = z3.is_true(model_value)
                    elif param_types[var_name] == "str":
                        str_val = str(model_value)
                        if str_val.startswith('"') and str_val.endswith('"'):
                            str_val = str_val[1:-1]
                        param_values[var_name] = str_val
                    else:
                        # Default to int for unrecognized types
                        param_values[var_name] = model_value.as_long()
                except Exception as e:
                    print(f"Couldn't get {var_name} from model: {e}")
                    # Use default values for parameters not in the model
                    if param_types[var_name] == "int":
                        param_values[var_name] = 0
                    elif param_types[var_name] == "float":
                        param_values[var_name] = 0.0
                    elif param_types[var_name] == "bool":
                        param_values[var_name] = False
                    elif param_types[var_name] == "str":
                        param_values[var_name] = ""
                    else:
                        param_values[var_name] = None
        
        return param_values
    
    def _order_parameters(self, func_node, param_values):
        ordered_params = []
        
        for arg in func_node.args.args:
            arg_name = arg.arg
            if arg_name == 'self':  # Skip self parameter
                continue
            if arg_name in param_values:
                ordered_params.append(param_values[arg_name])
            else:
                # Default value handling if parameter not in solution
                if arg.annotation and hasattr(arg.annotation, 'id'):
                    if arg.annotation.id == 'int':
                        ordered_params.append(0)
                    elif arg.annotation.id == 'float':
                        ordered_params.append(0.0)
                    elif arg.annotation.id == 'bool':
                        ordered_params.append(False)
                    elif arg.annotation.id == 'str':
                        ordered_params.append('')
                    else:
                        ordered_params.append(None)
                else:
                    ordered_params.append(None)
        
        return ordered_params
    
    def _order_arguments(self, func, args_dict):
        import inspect
        sig = inspect.signature(func)
        param_names = [p.name for p in sig.parameters.values() if p.name != 'self']
        
        ordered_args = []
        for name in param_names:
            if name in args_dict:
                ordered_args.append(args_dict[name])
            else:
                ordered_args.append(None)  # Default to None if missing
        
        return ordered_args
    
    def _add_random_test_case(self, function_metadata, test_cases):
        random_inputs = self.generate_random_inputs(function_metadata.params)
        try:
            module = self.analysis_context.module
            func_name = function_metadata.function_name
            
            if self._analysis_context.class_name:
                cls = getattr(module, self._analysis_context.class_name)
                obj = cls()
                func = getattr(obj, func_name)
            else:
                func = getattr(module, func_name)
            
            ordered_args = self._order_arguments(func, random_inputs)
            
            output = func(*ordered_args)
            test_cases.append(TestCase(func_name, tuple(ordered_args), output))
        except Exception as e:
            print(f"Error executing function with random inputs: {e}")
    """


class CLIView:
    def __init__(self):
        "pass"

    def display_message(self, message: str):
        print(f"[INFO] {message}")

    def display_error(self, error: str):
        print(f"[ERROR] {error}")

    def prompt_input(self, prompt: str) -> str:
        return input(f"{prompt}:> ")


import ast
import importlib.util
import os
import sys
from _ast import Module
from importlib import util
from types import ModuleType
from typing import Dict

def get_import_info(filepath: str) -> Dict[str, str]:
    if not os.path.exists(filepath) or not filepath.endswith('.py'):
        raise ValueError(f"Invalid Python file: {filepath}")
    
    # Get the directory and filename
    file_dir = os.path.dirname(os.path.abspath(filepath))
    module_name = os.path.splitext(os.path.basename(filepath))[0]
    
    # Check if this is part of a package (has __init__.py)
    is_package = os.path.exists(os.path.join(file_dir, '__init__.py'))
    
    # Find the project root by looking for setup.py or a .git directory
    project_root = find_project_root(file_dir)
    
    # Build the import path based on the file's location relative to the project root
    if project_root:
        rel_path = os.path.relpath(file_dir, project_root)
        if rel_path == '.':
            # File is directly in the project root
            import_path = module_name
            package_name = ''
        else:
            # File is in a subdirectory
            path_parts = rel_path.replace('\\', '/').split('/')
            # Filter out any empty parts
            path_parts = [part for part in path_parts if part]
            
            if path_parts:
                package_name = path_parts[0]
                # Construct the full import path
                import_path = '.'.join(path_parts) + '.' + module_name
            else:
                package_name = ''
                import_path = module_name
    else:
        # Fallback if we can't find a project root
        package_name = ''
        import_path = module_name
    
    info = {
        'module_name': module_name,
        'package_name': package_name,
        'import_path': import_path,
        'is_package': is_package,
        'project_root': project_root,
        'file_dir': file_dir
    }
    
    #print(f"INFO: {info}")
    
    return info

def find_project_root(start_dir: str) -> str | None:
    current_dir = start_dir
    
    # Walk up the directory tree
    while current_dir:
        # Check for common project root indicators
        if (os.path.exists(os.path.join(current_dir, 'setup.py')) or
            os.path.exists(os.path.join(current_dir, '.git')) or
            os.path.exists(os.path.join(current_dir, 'pyproject.toml'))):
            return current_dir
            
        if os.path.exists(os.path.join(current_dir, '__main__.py')):
            return current_dir
        
        # Check for 'testgen' directory which is your project name
        if os.path.basename(current_dir) == 'testgen':
            parent_dir = os.path.dirname(current_dir)
            return parent_dir
            
        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:  # Reached the root
            break
        current_dir = parent_dir
    
    return None


def load_module(file_path: str) -> ModuleType:
    # Load a Python module from a file path.
    if file_path is None:
        raise ValueError("File path not set! Use set_file_path() to specify the path of the file")

    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_filename(filepath: str) -> str:
    """Get filename from filepath."""
    return os.path.basename(filepath)


def load_and_parse_file_for_tree(file) -> Module:
    with open(file) as f:
        code = f.read()
    tree = ast.parse(code)
    return tree

def adjust_file_path_for_docker(file_path: str) -> str:
    """Adjust the file path to be valid inside the Docker container."""
    print(f"Docker - adjusting path: {file_path}")

    # If already absolute to /controller, return as-is
    if file_path.startswith("/controller/"):
        print(f"Docker - already adjusted: {file_path}")
        return file_path

    # If relative, make absolute
    adjusted_path = f"/controller/{file_path.lstrip('/')}"
    print(f"Docker - adjusted to: {adjusted_path}")
    return adjusted_path


"""def adjust_file_path_for_docker(file_path) -> str:
        #Adjust file path for Docker environment to handle subdirectories and relative paths.
        print(f"Docker - adjusting path: {file_path}")
        
        # Try direct path first (maybe it's correct already)
        if os.path.isfile(file_path):
            print(f"Docker - found file at direct path: {file_path}")
            return file_path
            
        # Try relative to /controller (root of mount)
        controller_path = os.path.join('/controller', file_path)
        if os.path.isfile(controller_path):
            print(f"Docker - found file at: {controller_path}")
            return controller_path
        
        # Try /controller/testgen/code_to_test path - this is where the file actually is
        testgen_path = os.path.join('/controller/testgen', file_path)
        if os.path.isfile(testgen_path):
            print(f"Docker - found file at: {testgen_path}")
            return testgen_path
        
        # If it's just a filename, search in common locations
        if os.path.basename(file_path) == file_path:
            for search_dir in ['/controller', '/controller/code_to_test', '/controller/testgen/code_to_test', '/controller/testgen']:
                test_path = os.path.join(search_dir, file_path)
                if os.path.isfile(test_path):
                    print(f"Docker - found file at: {test_path}")
                    return test_path
        
        # Debug output to help diagnose issues
        print("Docker - available files in /controller:")
        os.system("find /controller -name '*.py' | grep boolean")
        
        # Return original path if we couldn't find a better match
        print(f"Docker - couldn't find file, returning original: {file_path}")
        return file_path"""



def get_project_root_in_docker(script_path) -> str:
    # Don't use sys.argv[0] as it points to the virtualenv
    # Instead, find the actual project root
    if os.path.exists('/mnt/c/Users/cjsei/thesis/dev/testgen'):
        # Hard-coded path for now
        project_root = '/mnt/c/Users/cjsei/thesis/dev/testgen'
    else:
        # Try to find project root by looking for pyproject.toml or similar
        current_dir = os.path.dirname(os.path.abspath(script_path))
        while current_dir != '/':
            if os.path.exists(os.path.join(current_dir, 'pyproject.toml')) or \
               os.path.exists(os.path.join(current_dir, 'setup.py')):
                project_root = current_dir
                break
            current_dir = os.path.dirname(current_dir)
        else:
            # Fallback - use parent of script dir
            project_root = os.path.dirname(os.path.dirname(os.path.abspath(script_path)))
    
    print(f"Project root directory: {project_root}")
    return project_root